#include <apps/Common/wxAbstraction.h>
#include "LineProfileCanvas.h"
#include "PropViewFrame.h"
#include <wx/dcbuffer.h>

using namespace mvIMPACT::acquire;

IMPLEMENT_DYNAMIC_CLASS(LineProfileCanvas, DrawingCanvas)

BEGIN_EVENT_TABLE(LineProfileCanvas, DrawingCanvas)
  EVT_PAINT(LineProfileCanvas::OnPaint)
  EVT_SIZE(LineProfileCanvas::OnSize)
END_EVENT_TABLE()

//-----------------------------------------------------------------------------
LineProfileCanvas::LineProfileCanvas( wxWindow* parent, wxWindowID id /* = -1 */, const wxPoint& pos /* = wxDefaultPosition */,
				   const wxSize& size /* = wxDefaultSize */, long style /* = wxSUNKEN_BORDER */,
				   const wxString& name /* = "line profile" */, bool boActive /* = false */ )
	: ImageAnalysisPlotCanvas( parent, id, pos, size, style, name, boActive, 0 ), m_pHorData(0), m_pVerData(0)
//-----------------------------------------------------------------------------
{

}

//-----------------------------------------------------------------------------
LineProfileCanvas::~LineProfileCanvas()
//-----------------------------------------------------------------------------
{
	DeallocateProfileBuffer();
}

//-----------------------------------------------------------------------------
void LineProfileCanvas::DeallocateProfileBuffer( void )
//-----------------------------------------------------------------------------
{
	for( int channel=0; channel<m_ChannelCount; channel++ )
	{
		if( m_pHorData )
		{
			delete [] m_pHorData[channel];
		}
		if( m_pVerData )
		{
			delete [] m_pVerData[channel];
		}
	}
	delete [] m_pHorData;
	delete [] m_pVerData;
	m_pHorData = 0;
	m_pVerData = 0;
}

//-----------------------------------------------------------------------------
void LineProfileCanvas::DrawProfileLine( wxPaintDC& dc, int h, int startOffset, double scaleX, double scaleY, int** pData, int elementCount, TPlane plane )
//-----------------------------------------------------------------------------
{
	dc.SetPen( *m_Pens[plane].pColour_ );
	int lowerStart = h - GetBorderWidth();
	for( int i=0; i<elementCount-1; i++ )
	{
		dc.DrawLine( static_cast<int>( startOffset + ( i*scaleX ) ),
					 static_cast<int>( lowerStart - ( pData[plane][i] * scaleY ) ),
					 static_cast<int>( startOffset + ( ( i+1 ) * scaleX ) ),
					 static_cast<int>( lowerStart - ( pData[plane][i+1] * scaleY ) ) );
	}
	dc.SetPen( wxNullPen );
}

//-----------------------------------------------------------------------------
void LineProfileCanvas::OnPaint( wxPaintEvent & )
//-----------------------------------------------------------------------------
{
	wxCriticalSectionLocker locker( m_critSect );
	wxPaintDC dc( this );

	if( !IsActive() )
	{
		return;
	}

	wxCoord w, h;
	dc.GetSize(&w, &h);

	int borderWidth = GetBorderWidth();
	int diagramWidth = static_cast<int>(( w - ( 3 * borderWidth ) + 1.0 ) / 2.0 );
	int diagramHeight = h - ( 2 * borderWidth );
	int verDiagramStart = (2 * borderWidth ) + diagramWidth;
	double scaleY = static_cast<double>(diagramHeight) / static_cast<double>(m_valCount);

	dc.SetPen( *wxTRANSPARENT_PEN );
	dc.SetBrush( *wxWHITE_BRUSH );
	dc.DrawRectangle( borderWidth + 1, borderWidth, diagramWidth + 1, diagramHeight );
	dc.DrawRectangle( verDiagramStart + 1, borderWidth, diagramWidth + 1, diagramHeight );

	// hor. profile
	double scaleX = static_cast<double>(diagramWidth) / static_cast<double>( ( m_AOIw == 0 ) ? 1 : m_AOIw - 1);
	dc.SetPen( *wxBLACK );
	dc.DrawLine( borderWidth, h-borderWidth, borderWidth, borderWidth - 1 );
	dc.DrawLine( borderWidth, h-borderWidth, borderWidth + diagramWidth + 2, h-borderWidth );
	dc.DrawText( wxString::Format( wxT("%d"), m_AOIw ), borderWidth + diagramWidth, h - borderWidth + 3 );
	dc.DrawText( wxString::Format( wxT("%d"), m_valCount - 1 ), borderWidth/2, 1 );
	dc.DrawText( wxT("hor. profile"), diagramWidth/2, 1 );
	if( m_pHorData )
	{
		wxCoord xOffset(borderWidth);
		const wxCoord rectSize(10);
		wxCoord yOffset(h - borderWidth + 3);
		for( int channel=0; channel<m_ChannelCount; channel++ )
		{
			DrawProfileLine( dc, h, borderWidth + 1, scaleX, scaleY, m_pHorData, m_AOIw, static_cast<TPlane>(channel) );
			wxString completeDescString = wxString::Format( wxT("%s: "), m_Pens[channel].description_.c_str() );
			if( channel != 0 )
			{
				completeDescString.insert( 0, wxString(wxT(", ")) );
			}
			dc.DrawText( completeDescString, xOffset, yOffset );
			wxCoord textWidth, textHeight;
			dc.GetTextExtent( completeDescString, &textWidth, &textHeight );
			xOffset += textWidth;
			dc.SetPen( wxPen(*wxBLACK, 1, wxSOLID ) );
			dc.SetBrush( wxBrush(*(m_Pens[channel].pColour_)) );
			dc.DrawRectangle( xOffset, yOffset+3, rectSize, rectSize );
			dc.SetBrush( wxNullBrush );
			dc.SetPen( wxNullPen );
			xOffset += rectSize + 2;
		}
	}

	// ver. profile
	scaleX = static_cast<double>(diagramWidth) / static_cast<double>( ( m_AOIh == 0 ) ? 1 : m_AOIh - 1);
	dc.SetPen( *wxBLACK );
	dc.DrawLine( verDiagramStart, h-borderWidth, verDiagramStart, borderWidth - 1 );
	dc.DrawLine( verDiagramStart, h-borderWidth, verDiagramStart + diagramWidth + 2, h-borderWidth );
	dc.DrawText( wxString::Format( wxT("%d"), m_AOIh ), 2*(borderWidth + diagramWidth), h - borderWidth + 3 );
	dc.DrawText( wxString::Format( wxT("%d"), m_valCount - 1 ), static_cast<int>(1.5*static_cast<double>(borderWidth) + diagramWidth + 1), 1 );
	dc.DrawText( wxT("ver. profile"), static_cast<int>(borderWidth + ( 1.5 * static_cast<double>(diagramWidth) )), 1 );
	if( m_pVerData )
	{
		for( int channel=0; channel<m_ChannelCount; channel++ )
		{
			DrawProfileLine( dc, h, verDiagramStart + 1, scaleX, scaleY, m_pVerData, m_AOIh, static_cast<TPlane>(channel) );
		}
	}
	dc.SetBrush( wxNullBrush );
	dc.SetPen( wxNullPen );
}

//-----------------------------------------------------------------------------
void LineProfileCanvas::OnSize( wxSizeEvent& e )
//-----------------------------------------------------------------------------
{
	Refresh( true );
	e.Skip();
}

//-----------------------------------------------------------------------------
void LineProfileCanvas::RefreshData( const ImageBuffer* pIB, int x /* = -1 */, int y /* = -1 */, int w /* = -1 */, int h /* = -1 */, bool boForceRefresh /* = false */ )
//-----------------------------------------------------------------------------
{
	wxCriticalSectionLocker locker( m_critSect );

	bool boAOIChanged = false;
	bool boImageFormatChanged = false;
	if( !MustUpdate( pIB, x, y, w, h, boForceRefresh, &boAOIChanged, &boImageFormatChanged ) )
	{
		return;
	}

	if( boAOIChanged || !m_pHorData || !m_pVerData || boImageFormatChanged )
	{
		// free old memory
		DeallocateProfileBuffer();
		// allocate new memory
		m_ChannelCount = pIB->iChannelCount;
		m_pHorData = new int*[pIB->iChannelCount];
		m_pVerData = new int*[pIB->iChannelCount];
		for( int channel=0; channel<pIB->iChannelCount; channel++ )
		{
			m_pHorData[channel] = new int[m_AOIw];
			m_pVerData[channel] = new int[m_AOIh];
		}
	}

	assert( ( pIB->iChannelCount <= PLANE_CNT ) && wxT("A pixel format defines more channels then this tool can handle") );
	for( int channel=0; channel<pIB->iChannelCount; channel++ )
	{
		memset( m_pHorData[channel], 0, sizeof(int)*m_AOIw);
		memset( m_pVerData[channel], 0, sizeof(int)*m_AOIh);
		if( channel < PLANE_CNT )
		{
			m_Pens[channel].description_ = ConvertedString(pIB->pChannels[channel].szChannelDesc);
		}
	}

	switch( pIB->pixelFormat )
	{
	case ibpfMono10:
	case ibpfRGB101010Packed:
		m_valCount = VAL_COUNT_10_BIT;
		break;
	case ibpfRGB121212Packed:
	case ibpfMono12:
		m_valCount = VAL_COUNT_12_BIT;
		break;
	case ibpfRGB141414Packed:
	case ibpfMono14:
		m_valCount = VAL_COUNT_14_BIT;
		break;
	case ibpfRGB161616Packed:
	case ibpfMono16:
		m_valCount = VAL_COUNT_16_BIT;
		break;
	case ibpfRGBx888Packed:
	case ibpfRGB888Packed:
	case ibpfRGBx888Planar:
	case ibpfYUV422Packed:
	case ibpfYUV422Planar:
	case ibpfMono8:
	default:
		m_valCount = VAL_COUNT_8_BIT;
		break;
	}

	// calculate data
	switch( pIB->pixelFormat )
	{
	case ibpfMono8:
		for( int line=0; line<m_AOIh; line++ )
		{
			unsigned char* pData = static_cast<unsigned char*>(pIB->vpData) + ( ( m_AOIy + line ) * pIB->pChannels[pRed].iLinePitch ) + m_AOIx;
			for( int pixel=0; pixel<m_AOIw; pixel++ )
			{
				m_pHorData[pRed][pixel] += *pData;
				m_pVerData[pRed][line] += *pData++;
			}
		}
		break;
	case ibpfMono10:
	case ibpfMono12:
	case ibpfMono14:
	case ibpfMono16:
		for( int line=0; line<m_AOIh; line++ )
		{
			unsigned short* pData = static_cast<unsigned short*>(pIB->vpData) + ( ( m_AOIy + line ) * pIB->iWidth ) + m_AOIx;
			for( int pixel=0; pixel<m_AOIw; pixel++ )
			{
				m_pHorData[pRed][pixel] += saveAssign( *pData, static_cast<unsigned short>(0), static_cast<unsigned short>(m_valCount-1) );
				m_pVerData[pRed][line] += saveAssign( *pData, static_cast<unsigned short>(0), static_cast<unsigned short>(m_valCount-1) );
				++pData;
			}
		}
		break;
	case ibpfRGBx888Packed:
		for( int line=0; line<m_AOIh; line++ )
		{
			unsigned char* pData = static_cast<unsigned char*>(pIB->vpData) + ( ( m_AOIy + line ) * pIB->pChannels[pRed].iLinePitch ) + ( m_AOIx * pIB->iBytesPerPixel );
			for( int pixel=0; pixel<m_AOIw; pixel++ )
			{
				m_pHorData[pBlue][pixel] += *pData;
				m_pVerData[pBlue][line] += *pData++;
				m_pHorData[pGreen][pixel] += *pData;
				m_pVerData[pGreen][line] += *pData++;
				m_pHorData[pRed][pixel] += *pData;
				m_pVerData[pRed][line] += *pData++;
				*pData++;
			}
		}
		break;
	case ibpfRGB888Packed:
		for( int line=0; line<m_AOIh; line++ )
		{
			unsigned char* pData = static_cast<unsigned char*>(pIB->vpData) + ( ( m_AOIy + line ) * pIB->pChannels[pRed].iLinePitch ) + ( m_AOIx * pIB->iBytesPerPixel );
			for( int pixel=0; pixel<m_AOIw; pixel++ )
			{
				m_pHorData[pBlue][pixel] += *pData;
				m_pVerData[pBlue][line] += *pData++;
				m_pHorData[pGreen][pixel] += *pData;
				m_pVerData[pGreen][line] += *pData++;
				m_pHorData[pRed][pixel] += *pData;
				m_pVerData[pRed][line] += *pData++;
			}
		}
		break;
	case ibpfRGB101010Packed:
	case ibpfRGB121212Packed:
	case ibpfRGB141414Packed:
	case ibpfRGB161616Packed:
		for( int line=0; line<m_AOIh; line++ )
		{
			unsigned short* pData = reinterpret_cast<unsigned short*>(static_cast<unsigned char*>(pIB->vpData) + ( ( m_AOIy + line ) * pIB->pChannels[pRed].iLinePitch ) + ( m_AOIx * pIB->iBytesPerPixel ));
			for( int pixel=0; pixel<m_AOIw; pixel++ )
			{
				m_pHorData[pBlue][pixel] += saveAssign( *pData, static_cast<unsigned short>(0), static_cast<unsigned short>(m_valCount-1) );
				m_pVerData[pBlue][line] += saveAssign( *pData, static_cast<unsigned short>(0), static_cast<unsigned short>(m_valCount-1) );
				++pData;
				m_pHorData[pGreen][pixel] += saveAssign( *pData, static_cast<unsigned short>(0), static_cast<unsigned short>(m_valCount-1) );
				m_pVerData[pGreen][line] += saveAssign( *pData, static_cast<unsigned short>(0), static_cast<unsigned short>(m_valCount-1) );
				++pData;
				m_pHorData[pRed][pixel] += saveAssign( *pData, static_cast<unsigned short>(0), static_cast<unsigned short>(m_valCount-1) );
				m_pVerData[pRed][line] += saveAssign( *pData, static_cast<unsigned short>(0), static_cast<unsigned short>(m_valCount-1) );
				++pData;
			}
		}
		break;
	case ibpfRGBx888Planar:
		for( int line=0; line<m_AOIh; line++ )
		{
			unsigned char* pr = static_cast<unsigned char*>(pIB->vpData) + pIB->pChannels[pRed].iChannelOffset + ( ( m_AOIy + line ) * pIB->pChannels[pRed].iLinePitch ) + m_AOIx;
			unsigned char* pg = static_cast<unsigned char*>(pIB->vpData) + pIB->pChannels[pGreen].iChannelOffset + ( ( m_AOIy + line ) * pIB->pChannels[pGreen].iLinePitch ) + m_AOIx;
			unsigned char* pb = static_cast<unsigned char*>(pIB->vpData) + pIB->pChannels[pBlue].iChannelOffset + ( ( m_AOIy + line ) * pIB->pChannels[pBlue].iLinePitch ) + m_AOIx;
			for( int pixel=0; pixel<m_AOIw; pixel++ )
			{
				m_pHorData[pRed][pixel] += *pr;
				m_pVerData[pRed][line] += *pr++;
				m_pHorData[pGreen][pixel] += *pg;
				m_pVerData[pGreen][line] += *pg++;
				m_pHorData[pBlue][pixel] += *pb;
				m_pVerData[pBlue][line] += *pb++;
			}
		}
		break;
	case ibpfYUV422Packed:
		for( int line=0; line<m_AOIh; line++ )
		{
			unsigned char* pData = static_cast<unsigned char*>(pIB->vpData) + ( ( m_AOIy + line ) * pIB->pChannels[pRed].iLinePitch ) + ( m_AOIx * pIB->iBytesPerPixel );
			for( int pixel=0; pixel<m_AOIw; pixel++ )
			{
				m_pHorData[pY][pixel] += *pData;
				m_pVerData[pY][line] += *pData;
				if( ( m_AOIx + pixel ) & 1 )
				{
					m_pHorData[pV][pixel] += *(pData + 1);
					m_pVerData[pV][line] += *(pData + 1);
					m_pHorData[pU][pixel] += *(pData - 1);
					m_pVerData[pU][line] += *(pData - 1);
				}
				else
				{
					m_pHorData[pV][pixel] += *(pData - 1);
					m_pVerData[pV][line] += *(pData - 1);
					m_pHorData[pU][pixel] += *(pData + 1);
					m_pVerData[pU][line] += *(pData + 1);
				}
				pData += 2;
			}
		}
		break;
	case ibpfYUV422Planar:
		for( int line=0; line<m_AOIh; line++ )
		{
			unsigned char* py = static_cast<unsigned char*>(pIB->vpData) + pIB->pChannels[pRed].iChannelOffset + ( ( m_AOIy + line ) * pIB->pChannels[pY].iLinePitch ) + m_AOIx;
			unsigned char* pu = static_cast<unsigned char*>(pIB->vpData) + pIB->pChannels[pGreen].iChannelOffset + ( ( m_AOIy + line ) * pIB->pChannels[pU].iLinePitch ) + m_AOIx;
			unsigned char* pv = static_cast<unsigned char*>(pIB->vpData) + pIB->pChannels[pBlue].iChannelOffset + ( ( m_AOIy + line ) * pIB->pChannels[pV].iLinePitch ) + m_AOIx;
			for( int pixel=0; pixel<m_AOIw; pixel++ )
			{
				m_pHorData[pY][pixel] += *py;
				m_pVerData[pY][line] += *py++;
				m_pHorData[pU][pixel] += *pu;
				m_pVerData[pU][line] += *pu;
				m_pHorData[pV][pixel] += *pv;
				m_pVerData[pV][line] += *pv;
				if( pixel & 1 )
				{
					++pu;
					++pv;
				}
			}
		}
		break;
	default:
		// unsupported colour format
		break;
	}

	for( int channel=0; channel<m_ChannelCount; channel++ )
	{
		for( int i=0; i<m_AOIw; i++ )
		{
			m_pHorData[channel][i] /= m_AOIh;
		}
		for( int i=0; i<m_AOIh; i++ )
		{
			m_pVerData[channel][i] /= m_AOIw;
		}
	}

	bool boFullUpdate = boAOIChanged || boImageFormatChanged;
	Refresh(boFullUpdate);
}
